/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File fdi.c - VL 1772 disk controller emulation for SimCoupe
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#ifdef linux
#include <sys/file.h>
#include <sys/ioctl.h>
#include <linux/fd.h>
#endif
#include <unistd.h>

#include "z80.h"
#include "sam.h"
#include "util.h"
#include "fdi.h"

typedef struct _vl1772_reg {
		uns8	Command;
		uns8	Status;
		uns8	Track;
		uns8	Sector;
		uns8	Data;
} vl1772_reg;

typedef struct _DiskAttr {
		uns8	NoSides;
		uns8	NoTracks;
		uns8	NoSectors;
} DiskAttr;

typedef struct _Aside	{
		char 		*FileName;
		DiskAttr        Dattr;
		vl1772_reg	reg;
		uns8		SectorBuf[SECTOR_SIZE];
		uns8		*Bufp;
		int		ByteCount;
} ASide;

#define D_DISK(x)	(((x) & 0x10) >> 4)
#define D_SIDE(x)	(((x) & 0x04) >> 2)
#define DISK(x)		TheDisks[D_DISK((x) & 0xff)][D_SIDE((x) & 0xff)]

enum {COMMAND = 0, STATUS = 0, TRACK, SECTOR, DATA};


static ASide	TheDisks[2][MAX_NO_SIDES];


int ReadSector (uns8 Side_no, ASide *Dside);
int WriteSector(uns8 Side_no, ASide *Dside);

void 
fdi_out(uns16 port, uns8 byte)
{
	int Sdir = 1;

	switch (port & 0x03) {

		case COMMAND:
/*                        printf("Command : %x\n",(byte & 0xf0)); */
			DISK(port).reg.Command = byte & 0xf0;

			switch (byte & 0xf0) {
			/* Type I commands */
				case RESTORE:
					DISK(port).reg.Track = 0;
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP | LOST_DATA;
					break;
				case SEEK:
					DISK(port).reg.Track = DISK(port).reg.Data;
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					if (!DISK(port).reg.Track) 
					   DISK(port).reg.Status |= LOST_DATA;
					break;
				case STEP_NUPD:
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					break;
				case STEP_UPD:
					DISK(port).reg.Track += Sdir;
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					if (!DISK(port).reg.Track) 
					   DISK(port).reg.Status |= LOST_DATA;
					break;
				case STEP_IN_NUPD:
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					break;
				case STEP_IN_UPD:
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					DISK(port).reg.Track++;
					break;
				case STEP_OUT_NUPD:
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					break;
				case STEP_OUT_UPD:
					DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
					if (!DISK(port).reg.Track--) 
					   DISK(port).reg.Status |= LOST_DATA;
					break;
				/* Type II Commands */
				case READ_1SECTOR:
				case READ_MSECTOR:
				 DISK(port).reg.Status = BUSY | DRQ;
               			 if (ReadSector(D_SIDE(port),&DISK(port))) {
					DISK(port).reg.Status = RECORD_NOT_FOUND;
				 }
				 break;
				case WRITE_1SECTOR:
				case WRITE_MSECTOR:
				 DISK(port).reg.Status = BUSY | DRQ;
				 DISK(port).Bufp = DISK(port).SectorBuf;
				 DISK(port).ByteCount = 0;
				 break;
				/* Type III Commands */
				case READ_ADDRESS:
				case READ_TRACK:
				case WRITE_TRACK:
					break;
				/* Type IV Commands */
				case FORCE_INTERRUPT:
					DISK(port).reg.Status &= ~BUSY;
					DISK(port).reg.Command = 0;
					break;
			
				default:
					printf("Unknown Disk Command %x\n",byte & 0xf0);
					break;
			} /* End switch */
			break;

		case TRACK:
			DISK(port).reg.Track = byte;
			break;
		case SECTOR:
			DISK(port).reg.Sector = byte;
			break;
		case DATA:
			if (DISK(port).reg.Status & BUSY) {
			 if (DISK(port).reg.Command == WRITE_1SECTOR ||
			     DISK(port).reg.Command == WRITE_MSECTOR) {
				*DISK(port).Bufp++ = byte;
/*                                printf("ByteCount = %d\n",DISK(port).ByteCount);*/
				if (++DISK(port).ByteCount == SECTOR_SIZE) {
				  if (WriteSector(D_SIDE(port),&DISK(port))) {
					DISK(port).reg.Status = RECORD_NOT_FOUND;
					break;
				  }
				  if (DISK(port).reg.Command == READ_MSECTOR &&
				      DISK(port).reg.Sector++ != DISK(port).Dattr.NoSectors) {
					DISK(port).Bufp = DISK(port).SectorBuf;
					DISK(port).ByteCount = 0;
			          } else {
				   DISK(port).reg.Status &= ~(BUSY | DRQ);
				   DISK(port).reg.Command = 0;
				  }					
				}
			 }	
			}	
			DISK(port).reg.Data = byte;
			break;
	} /* End Switch */


}

uns8 
fdi_in(uns16 port)
{
	uns8 byte;

	static uns8 DRQ_count = 7;
	
	switch (port & 0x03) {

		case STATUS :
/*                        printf("Status : %x\n", DISK(port).reg.Status); */
			byte = DISK(port).reg.Status;
			if ((!DISK(port).reg.Command) && !DRQ_count--) {
				DISK(port).reg.Status ^= DRQ;
				DRQ_count = 7;
			}
			break;
		case TRACK:
			byte = DISK(port).reg.Track;
			break;
		case SECTOR:
			byte = DISK(port).reg.Sector;
			break;
		case DATA:
/*                        printf("Data : %x\n",DISK(port).ByteCount); */
                        if (DISK(port).reg.Status & BUSY) {
			 if (DISK(port).reg.Command == READ_1SECTOR ||
			     DISK(port).reg.Command == READ_MSECTOR) {
				DISK(port).reg.Data = *DISK(port).Bufp++;
				if (++DISK(port).ByteCount == SECTOR_SIZE) {
				 if (DISK(port).reg.Command == READ_MSECTOR &&
				     DISK(port).reg.Sector++ != DISK(port).Dattr.NoSectors) {
				  if (ReadSector(D_SIDE(port),&DISK(port))) {
					DISK(port).reg.Status = RECORD_NOT_FOUND;
					break;
				  }
			         } else {
				  DISK(port).reg.Status &= ~(BUSY | DRQ);
       			          DISK(port).reg.Command = 0;
				 }					
				}
			 }	
			}	
			byte = DISK(port).reg.Data;
			break;
	} /* End Switch */

	return(byte);
}

void 
insert_disk(char *fn, int disk_no)
{

	struct stat stbuff;
	DiskAttr    tmpAttr;

	disk_no--;

	/* Try to work out the type of disk image inserted */

	if (!stat(fn, &stbuff)) {

		if ((stbuff.st_mode & S_IFMT) == S_IFREG) {
			/* Regular file */
			switch (stbuff.st_size) {
				case SAM_FORMAT_SIZE:
					simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): SAM type image : %s\n",fn);
					tmpAttr.NoSides = 2;
					tmpAttr.NoTracks = 80;
					tmpAttr.NoSectors = 10;
					break;
				case MSDOS_FORMAT_SIZE:
					simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): MSDOS/CPM type image : %s",fn);
					tmpAttr.NoSides = 2;
					tmpAttr.NoTracks = 80;
					tmpAttr.NoSectors = 9;
					break;
				default:
					simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Unknown image type : %s",fn);
					return;
					break;
			} /* End switch */
		}
		else if (((stbuff.st_mode & S_IFMT) == S_IFBLK)
#ifdef linux
		         && ( stbuff.st_rdev/256 == 2))  {
				/* For Linux Major device 2 is a floppy drive */
				int fd;
				struct floppy_struct fstruct;

				fd = open(fn, O_RDONLY, 0);
				if (!ioctl(fd, FDGETPRM, &fstruct)) {
					tmpAttr.NoSides   = fstruct.head;
					tmpAttr.NoTracks  = fstruct.track;
					tmpAttr.NoSectors = fstruct.sect;
					close(fd);
				}
				else {
					simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Cannot get floppy parameters : %s\n",fn);
					simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Assuming Sam Disk\n");
					tmpAttr.NoSides = 2;
					tmpAttr.NoTracks = 80;
					tmpAttr.NoSectors = 10;
					close(fd);
				}	
#else
			) {
			simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Direct access to floppy disk only"
						" available on Linux systems\n");
			return;
#endif 
		}
		else {
			simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Wrong file type : %s\n",fn);
			return;			
		}
		if (tmpAttr.NoSectors > 10) {
			simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Cannot use disks with more than 10"
						"sectors/track\n : %s\n",fn);
			return;
		}

		TheDisks[disk_no][0].FileName = malloc(strlen(fn));
		strcpy(TheDisks[disk_no][0].FileName, fn);
		TheDisks[disk_no][1].FileName = TheDisks[disk_no][0].FileName;

		TheDisks[disk_no][0].Dattr = tmpAttr;
		TheDisks[disk_no][1].Dattr = tmpAttr;
		

	}
	else {
		simcoupe_mesg(SIMCOUPE_WARN,"insert_disk() : cannot stat %s\n",fn);		
	}
}

void 
eject_disk(int disk_no)
{

	disk_no--;
  
	free(TheDisks[disk_no][0].FileName);
	TheDisks[disk_no][0].FileName = NULL;
	TheDisks[disk_no][1].FileName = NULL;

}

int 
ReadSector(uns8 Side_no, ASide *Dside)
{
	FILE *fp;
	long offset;

        int i;
        
	if ((fp = open_simcoupe_file(Dside->FileName, "rb")) == NULL) {
		return -1; /* couldn't open image => drive not ready */
	}
	else {

	 offset = (SECTOR_SIZE * Dside->Dattr.NoSectors) * (2*Dside->reg.Track + Side_no) +
	          SECTOR_SIZE * (  Dside->reg.Sector - 1);
/*         printf("Side = %d,  Track = %d, Sector = %d, Offset = %d\n", 
                 Side_no, Dside->reg.Track, Dside->reg.Sector, offset);
*/	
	           	

         if (fseek(fp, offset, SEEK_SET) != 0) {
                 fclose(fp);
                 return -2; /* couldn't seek the required sector */ 	 
         }
         else {
         

          if (fread(Dside->SectorBuf, sizeof(uns8), SECTOR_SIZE, fp) != SECTOR_SIZE) {
                  fclose(fp);
                  return -3; /* couldn't read the required sector */
          }         
          else {

	  	fclose (fp);
/*
                for (i = 0; i < 512 ; i++) 
                    printf("%x",Dside->SectorBuf[i]);
                printf ("\n");
*/
		Dside -> ByteCount = 0;
		Dside -> Bufp = Dside -> SectorBuf;
		return 0;	
	  }
	 }
	}
}

int 
WriteSector(uns8 Side_no, ASide *Dside)
{
	FILE *fp;	
	long offset;
	
	if ((fp = open_simcoupe_file(Dside->FileName, "r+b")) == NULL) {
                fclose(fp);
		return -1; /* couldn't open image => drive not ready */
	}
	else {
	
	 offset = (SECTOR_SIZE * Dside->Dattr.NoSectors) * (2*Dside->reg.Track + Side_no) +
	          SECTOR_SIZE * (  Dside->reg.Sector - 1);
/*         printf("Side = %d,  Track = %d, Sector = %d, Offset = %d\n", 
                 Side_no, Dside->reg.Track, Dside->reg.Sector, offset);*/
	           	

         if (fseek(fp, offset, SEEK_SET) != 0) {
                 fclose(fp);
                 return -2; /* couldn't seek the required sector */ 	 
         }
         else {
         
          if (fwrite(Dside->SectorBuf, sizeof(uns8), SECTOR_SIZE, fp) != SECTOR_SIZE) {
                  fclose(fp);
                  return -3; /* couldn't read the required sector */
          }         
          else {
	
	 	fflush(fp);
		fclose (fp);
		Dside -> ByteCount = 0;
		Dside -> Bufp = Dside -> SectorBuf;
		return 0;
	   }
	  }
	 }
}

